@@ -17,61 +17,45 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* }}} */
 
 #include <config.h>
+
+#ifdef WITH_BACKGROUND
+
 #include <stdarg.h>
 #include <sys/types.h>
-#ifdef HAVE_SYS_WAIT_H
-#    include <sys/wait.h>
-#endif
 #include <errno.h>
 #ifdef HAVE_UNISTD_H
 #    include <unistd.h>
 #endif
 #include <sys/stat.h>
-#include <sys/param.h>
+#include <signal.h>	/* kill() */
 #include <string.h>
-#include <fcntl.h>
 #include <stdio.h>
-#include "tty.h"
-#include "global.h"
-#include "dlg.h"
-#include "widget.h"
-#include "wtools.h"
 
-/*
- * We currenlty only support one way of comunicating the background
- * and foreground process by using the socketpair system call
- */
-#ifdef USE_NETCODE
-#    include <sys/socket.h>
-#endif
-#include "dialog.h"
-#include "fileopctx.h"
-#include "key.h"	/* For add_select_channel(), delete_select_channel() */
-#include "eregex.h"
-#include "file.h"
-#include "filegui.h"
+#include "global.h"
+#include "background.h"
+#include "tty.h"	/* doupdate() */
+#include "dialog.h"	/* do_refresh() */
+#include "wtools.h"
+#include "fileopctx.h"	/* FileOpContext */
+#include "key.h"	/* add_select_channel(), delete_select_channel() */
+
+enum ReturnType {
+    Return_String,
+    Return_Integer
+};
 
 /* If true, this is a background process */
 int we_are_background = 0;
 
-#ifdef WITH_BACKGROUND
-/* If set background tasks wait to be attached */
-int background_wait = 0;
-
-#ifndef HAVE_SOCKETPAIR
-int socketpair(int, int, int, int fd[2]);
-#endif
-
 /* File descriptor for talking to our parent */
 static int parent_fd;
 
 #define MAXCALLARGS 4		/* Number of arguments supported */
-#define mymsg "Desde el hijo\n\r"
 
 struct TaskList *task_list = NULL;
 
@@ -124,101 +108,50 @@ unregister_task_running (pid_t pid, int fd)
  * -1 on failure
  */
 int
-do_background (FileOpContext *ctx, char *info)
+do_background (struct FileOpContext *ctx, char *info)
 {
-    int comm [2];		/* control connection stream */
+    int comm[2];		/* control connection stream */
     pid_t pid;
 
-    if (socketpair (AF_UNIX, SOCK_STREAM, 0, comm) == -1)
+    if (pipe (comm) == -1)
 	return -1;
 
-    if ((pid = fork ()) == -1)
+    if ((pid = fork ()) == -1) {
+    	int saved_errno = errno;
+    	(void) close (comm[0]);
+    	(void) close (comm[1]);
+    	errno = saved_errno;
 	return -1;
+    }
 
-    if (pid == 0){
+    if (pid == 0) {
 	int nullfd;
 
-	parent_fd = comm [1];
+	close (comm[0]);
+	parent_fd = comm[1];
 	we_are_background = 1;
+	current_dlg = NULL;
 
 	/* Make stdin/stdout/stderr point somewhere */
 	close (0);
 	close (1);
 	close (2);
 
-	if ((nullfd = open ("/dev/null", O_RDONLY)) != -1){
-	    while (dup2 (nullfd, 0) == -1 && errno == EINTR)
-		    ;
-	    while (dup2 (nullfd, 1) == -1 && errno == EINTR)
-		    ;
-	    while (dup2 (nullfd, 2) == -1 && errno == EINTR)
-		    ;
+	if ((nullfd = open ("/dev/null", O_RDWR)) != -1) {
+	    while (dup2 (nullfd, 0) == -1 && errno == EINTR);
+	    while (dup2 (nullfd, 1) == -1 && errno == EINTR);
+	    while (dup2 (nullfd, 2) == -1 && errno == EINTR);
 	}
 
-	/* To make it obvious if it fails, there is a bug report on this */
-	write (2, mymsg, sizeof (mymsg));
-	write (1, mymsg, sizeof (mymsg));
-
-	/* Just for debugging the background back end */
-	if (background_wait){
-	    volatile int i = 1;
-	    
-	    while (i)
-		;
-	}
 	return 0;
     } else {
-	close (comm [1]);
+	close (comm[1]);
 	ctx->pid = pid;
-	register_task_running (ctx, pid, comm [0], info);
+	register_task_running (ctx, pid, comm[0], info);
 	return 1;
     }
 }
 
-static char *
-background_title (char *str)
-{
-    return g_strconcat (_("Background process:"), str, NULL);
-}
-
-/* {{{ Routines that do the real job */
-static void
-real_message_1s (enum OperationMode mode, int *flags, char *title, char *str1)
-{
-    if (mode == Background)
-	title = background_title (title);
-    
-    message (*flags, title, str1);
-
-    if (mode == Background)
-	g_free (title);
-}
-
-static void
-real_message_2s (enum OperationMode mode, int *flags, char *title, char *str1, char *str2)
-{
-    if (mode == Background)
-	title = background_title (title);
-    
-    message (*flags, title, str1, str2);
-    
-    if (mode == Background)
-	g_free (title);
-}
-
-static void
-real_message_3s (enum OperationMode mode, int *flags, char *title, char *str1, char *str2, const char *str3)
-{
-    if (mode == Background)
-	title = background_title (title);
-    
-    message (*flags, title, str1, str2, str3);
-    
-    if (mode == Background)
-	g_free (title);
-}
-/* }}} */
-
 /* {{{ Parent handlers */
 
 /* Parent/child protocol
@@ -248,10 +181,10 @@ real_message_3s (enum OperationMode mode, int *flags, char *title, char *str1, c
  *
  * If the return type is a string:
  *
- *     the parent writes the resulting string lenght
+ *     the parent writes the resulting string length
  *     if the result string was NULL or the empty string,
- *     then the lenght is zero.
- *     The parent then writes the string lenght and frees
+ *     then the length is zero.
+ *     The parent then writes the string length and frees
  *     the result string.
  */
 /*
@@ -267,28 +200,29 @@ background_attention (int fd, void *closure)
     void *routine;
     int  argc, i, result, status;
     char *data [MAXCALLARGS];
-    int bytes;
+    ssize_t bytes;
     enum ReturnType type;
 
     ctx = closure;
 
     bytes = read (fd, &routine, sizeof (routine));
-    if (bytes < (sizeof (routine))){
-	char *background_process_error = _(" Background process error ");
+    if (bytes == -1 || (size_t) bytes < (sizeof (routine))) {
+	const char *background_process_error = _(" Background process error ");
 
-	if (errno == ECHILD)
-	    message (1, background_process_error, _(" Child died unexpectedly "));
-	else
+	unregister_task_running (ctx->pid, fd);
+	if (!waitpid (ctx->pid, &status, WNOHANG)) {
+	    /* the process is still running, but it misbehaves - kill it */
+	    kill (ctx->pid, SIGTERM);
 	    message (1, background_process_error, _(" Unknown error in child "));
-	unregister_task_running (ctx->pid, fd);
-	waitpid (ctx->pid, &status, 0);
-	return 0;
-    }
+	    return 0;
+	}
+
+	/* 0 means happy end */
+	if (WIFEXITED (status) && (WEXITSTATUS (status) == 0))
+	    return 0;
+
+	message (1, background_process_error, _(" Child died unexpectedly "));
 
-    /* If the routine is zero, then the child is telling us that he is dying */
-    if ((int) routine == MSG_CHILD_EXITING){
-	unregister_task_running (ctx->pid, fd);
-	waitpid (ctx->pid, &status, 0);
 	return 0;
     }
 
@@ -315,7 +249,7 @@ background_attention (int fd, void *closure)
 
     /* Handle the call */
     if (type == Return_Integer){
-	if (have_ctx)
+	if (!have_ctx)
 	    switch (argc){
 	    case 1:
 		result = (*(int (*)(int, char *))routine)(Background, data [0]);
@@ -399,9 +333,7 @@ background_attention (int fd, void *closure)
 
     do_refresh ();
     mc_refresh ();
-#ifndef HAVE_X
     doupdate ();
-#endif
     return 0;
 }
 
@@ -431,7 +363,7 @@ parent_call_header (void *routine, int argc, enum ReturnType type, FileOpContext
 }
 
 int
-parent_call (void *routine, FileOpContext *ctx, int argc, ...)
+parent_call (void *routine, struct FileOpContext *ctx, int argc, ...)
 {
     va_list ap;
     int i;
@@ -454,7 +386,7 @@ parent_call (void *routine, FileOpContext *ctx, int argc, ...)
     return i;
 }
 
-static char *
+char *
 parent_call_string (void *routine, int argc, ...)
 {
     va_list ap;
@@ -481,95 +413,4 @@ parent_call_string (void *routine, int argc, ...)
     return str;
 }
 
-void
-tell_parent (int msg)
-{
-    write (parent_fd, &msg, sizeof (int));
-}
-
-void
-message_1s (int flags, char *title, char *str1)
-{
-    if (we_are_background)
-	parent_call ((void *)real_message_1s, NULL, 3, sizeof (flags), &flags,
-		     strlen (title), title, strlen (str1), str1);
-    else
-	real_message_1s (Foreground, &flags, title, str1);
-}
-
-void
-message_2s (int flags, char *title, char *str1, char *str2)
-{
-    if (we_are_background)
-	parent_call ((void *)real_message_2s, NULL, 4, sizeof (flags), &flags,
-		     strlen (title), title, strlen (str1), str1,
-	             strlen (str2), str2);
-    else
-	real_message_2s (Foreground, &flags, title, str1, str2);
-}
-
-void
-message_3s (int flags, char *title, char *str1, char *str2, const char *str3)
-{
-    if (we_are_background)
-	parent_call ((void *)real_message_3s, NULL, 3, sizeof (flags), &flags,
-		     strlen (title), title, strlen (str1), str1,
-	             strlen (str2), str2, strlen (str3), str3);
-    else
-	real_message_3s (Foreground, &flags, title, str1, str2, str3);
-}
-
-char *
-input_dialog_help (char *header, char *text, char *help, char *def_text)
-{
-    if (we_are_background)
-	return parent_call_string ((void *)real_input_dialog_help, 4,
-				   strlen (header),   header,
-				   strlen (text),     text,
-				   strlen (help),     help,
-				   strlen (def_text), def_text);
-    else
-	return real_input_dialog_help (header, text, help, def_text);
-}
-
-#else /* Else => No background code support */
-
-/* {{{ Stubs if background code is not supported */
-void
-message_1s (int flags, char *title, char *str1)
-{
-    message (flags, title, str1);
-}
-
-void
-message_2s (int flags, char *title, char *str1, char *str2)
-{
-    message (flags, title, str1, str2);
-}
-
-void
-message_3s (int flags, char *title, char *str1, char *str2, const char *str3)
-{
-    message (flags, title, str1, str2, str3);
-}
-
-char *
-input_dialog_help (char *header, char *text, char *help, char *def_text)
-{
-    return real_input_dialog_help (header, text, help, def_text);
-}
-/* }}} */
-
-#endif
-
-/* {{{ Functions shared between background and foreground */
-
-void
-message_1s1d (int flags, char *title, char *str, int d)
-{
-    char *p = g_strdup_printf (str, d);
-    message_1s (flags, title, p);
-    g_free (p);
-}
-
-/* }}} */
+#endif				/* WITH_BACKGROUND */
