@@ -14,7 +14,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
    Implements the hypertext file viewer.
    The hypertext file is a file that may have one or more nodes.  Each
@@ -45,12 +45,11 @@
 #include "global.h"
 #include "tty.h"
 #include "color.h"
-#include "dialog.h"
 #include "win.h"
 #include "mouse.h"
-#include "key.h"	/* For mi_getch() */
+#include "key.h"		/* For mi_getch() */
 #include "help.h"
-#include "dlg.h"		/* For Dlg_head */
+#include "dialog.h"		/* For Dlg_head */
 #include "widget.h"		/* For Widget */
 #include "wtools.h"		/* For common_dialog_repaint() */
 
@@ -66,31 +65,31 @@
 static char *data;		/* Pointer to the loaded data file */
 static int help_lines;		/* Lines in help viewer */
 static int  history_ptr;	/* For the history queue */
-static char *main_node;		/* The main node */
-static char *last_shown = 0;	/* Last byte shown in a screen */
+static const char *main_node;	/* The main node */
+static const char *last_shown = NULL;	/* Last byte shown in a screen */
 static int end_of_node = 0;	/* Flag: the last character of the node shown? */
-static char *currentpoint, *startpoint;
-static char *selected_item;
+static const char *currentpoint, *startpoint;
+static const char *selected_item;
 
 /* The widget variables */
 static Dlg_head *whelp;
 
 static struct {
-    char *page;			/* Pointer to the selected page */
-    char *link;			/* Pointer to the selected link */
+    const char *page;		/* Pointer to the selected page */
+    const char *link;		/* Pointer to the selected link */
 } history [HISTORY_SIZE];
 
 /* Link areas for the mouse */
 typedef struct Link_Area {
     int x1, y1, x2, y2;
-    char *link_name;
+    const char *link_name;
     struct Link_Area *next;
 } Link_Area;
 
 static Link_Area *link_area = NULL;
 static int inside_link_area = 0;
 
-static int help_callback (struct Dlg_head *h, int id, int msg);
+static cb_ret_t help_callback (struct Dlg_head *h, dlg_msg_t, int parm);
 
 #ifdef HAS_ACS_AS_PCCHARS
 static const struct {
@@ -125,11 +124,13 @@ static int acs2pc (int acscode)
 
 /* returns the position where text was found in the start buffer */
 /* or 0 if not found */
-static char *
-search_string (char *start, char *text)
+static const char *
+search_string (const char *start, const char *text)
 {
-    char *d = text;
-    char *e = start;
+    const char *result = NULL;
+    char *local_text = g_strdup (text);
+    char *d = local_text;
+    const char *e = start;
 
     /* fmt sometimes replaces a space with a newline in the help file */
     /* Replace the newlines in the link name with spaces to correct the situation */
@@ -139,23 +140,27 @@ search_string (char *start, char *text)
 	d++;
     }
     /* Do search */
-    for (d = text; *e; e++){
+    for (d = local_text; *e; e++){
 	if (*d == *e)
 	    d++;
 	else
-	    d = text;
-	if (!*d)
-	    return e+1;
+	    d = local_text;
+	if (!*d) {
+	    result = e + 1;
+	    goto cleanup;
+	}
     }
-    return 0;
+cleanup:
+    g_free (local_text);
+    return result;
 }
 
 /* Searches text in the buffer pointed by start.  Search ends */
 /* if the CHAR_NODE_END is found in the text.  Returns 0 on failure */
-static char *search_string_node (char *start, char *text)
+static const char *search_string_node (const char *start, const char *text)
 {
-    char *d = text;
-    char *e = start;
+    const char *d = text;
+    const char *e = start;
 
     if (!start)
 	return 0;
@@ -173,9 +178,9 @@ static char *search_string_node (char *start, char *text)
 
 /* Searches the_char in the buffer pointer by start and searches */
 /* it can search forward (direction = 1) or backward (direction = -1) */
-static char *search_char_node (char *start, char the_char, int direction)
+static const char *search_char_node (const char *start, char the_char, int direction)
 {
-    char *e;
+    const char *e;
 
     e = start;
     
@@ -187,9 +192,9 @@ static char *search_char_node (char *start, char the_char, int direction)
 }
 
 /* Returns the new current pointer when moved lines lines */
-static char *move_forward2 (char *c, int lines)
+static const char *move_forward2 (const char *c, int lines)
 {
-    char *p;
+    const char *p;
     int  line;
 
     currentpoint = c;
@@ -202,9 +207,9 @@ static char *move_forward2 (char *c, int lines)
     return currentpoint = c;
 }
 
-static char *move_backward2 (char *c, int lines)
+static const char *move_backward2 (const char *c, int lines)
 {
-    char *p;
+    const char *p;
     int line;
 
     currentpoint = c;
@@ -254,10 +259,10 @@ static void move_to_bottom (int dummy)
     move_backward (help_lines - 1);
 }
 
-static char *help_follow_link (char *start, char *selected_item)
+static const char *help_follow_link (const char *start, const char *selected_item)
 {
     char link_name [MAXLINKNAME];
-    char *p;
+    const char *p;
     int  i = 0;
 
     if (!selected_item)
@@ -280,9 +285,9 @@ static char *help_follow_link (char *start, char *selected_item)
     return _(" Help file format error\n");
 }
 
-static char *select_next_link (char *start, char *current_link)
+static const char *select_next_link (const char *start, const char *current_link)
 {
-    char *p;
+    const char *p;
 
     if (!current_link)
 	return 0;
@@ -296,9 +301,9 @@ static char *select_next_link (char *start, char *current_link)
     return p - 1;
 }
 
-static char *select_prev_link (char *start, char *current_link)
+static const char *select_prev_link (const char *start, const char *current_link)
 {
-    char *p;
+    const char *p;
 
     if (!current_link)
 	return 0;
@@ -311,7 +316,7 @@ static char *select_prev_link (char *start, char *current_link)
     return p;
 }
 
-static void start_link_area (int x, int y, char *link_name)
+static void start_link_area (int x, int y, const char *link_name)
 {
     Link_Area *new;
 
@@ -356,9 +361,9 @@ static void clear_link_areas (void)
     inside_link_area = 0;
 }
 
-static void help_show (Dlg_head *h, char *paint_start)
+static void help_show (Dlg_head *h, const char *paint_start)
 {
-    char *p;
+    const char *p;
     int  col, line, c;
     int  painting = 1;
     int acs;			/* Flag: Alternate character set active? */
@@ -437,15 +442,11 @@ static void help_show (Dlg_head *h, char *paint_start)
 		    if (c == ' ' || c == '.')
 			addch (c);
 		    else
-#ifndef NATIVE_WIN32
 #ifndef HAVE_SLANG
 			addch (acs_map [c]);
 #else
 			SLsmg_draw_object (h->y + line + 2, h->x + col + 2, c);
 #endif
-#else
-			addch (acs2pc (c));
-#endif /* NATIVE_WIN32 */
 		} else
 		    addch (c);
 		col++;
@@ -478,8 +479,9 @@ static int help_event (Gpm_Event *event, Widget *w)
 	return 0;
 
     /* The event is relative to the dialog window, adjust it: */
+    event->x -= 2;
     event->y -= 2;
-    
+
     if (event->buttons & GPM_B_RIGHT){
 	currentpoint = startpoint = history [history_ptr].page;
 	selected_item = history [history_ptr].link;
@@ -487,7 +489,7 @@ static int help_event (Gpm_Event *event, Widget *w)
 	if (history_ptr < 0)
 	    history_ptr = HISTORY_SIZE-1;
 	
-	help_callback (w->parent, 0, DLG_DRAW);
+	help_callback (w->parent, DLG_DRAW, 0);
 	return 0;
     }
 
@@ -533,7 +535,7 @@ static int help_event (Gpm_Event *event, Widget *w)
     }
 
     /* Show the new node */
-    help_callback (w->parent, 0, DLG_DRAW);
+    help_callback (w->parent, DLG_DRAW, 0);
 
     return 0;
 }
@@ -547,13 +549,13 @@ help_help_cmd (Dlg_head *h)
     history [history_ptr].link = selected_item;
     currentpoint = startpoint = search_string (data, "[How to use help]") + 1;
     selected_item = NULL;
-    help_callback (h, 0, DLG_DRAW);
+    help_callback (h, DLG_DRAW, 0);
 }
 
 static void
 help_index_cmd (Dlg_head * h)
 {
-    char *new_item;
+    const char *new_item;
 
     if (!(new_item = search_string (data, "[Contents]"))) {
 	message (1, MSG_ERROR, _(" Cannot find node %s in help file "),
@@ -567,7 +569,7 @@ help_index_cmd (Dlg_head * h)
 
     currentpoint = startpoint = new_item + 1;
     selected_item = NULL;
-    help_callback (h, 0, DLG_DRAW);
+    help_callback (h, DLG_DRAW, 0);
 }
 
 static void help_quit_cmd (void *x)
@@ -583,30 +585,33 @@ static void prev_node_cmd (Dlg_head *h)
     if (history_ptr < 0)
 	history_ptr = HISTORY_SIZE-1;
     
-    help_callback (h, 0, DLG_DRAW);
+    help_callback (h, DLG_DRAW, 0);
 }
 
-static int md_callback (Widget *w, int msg, int par)
+static cb_ret_t
+md_callback (Widget *w, widget_msg_t msg, int parm)
 {
-    return default_proc (msg, par);
+    return default_proc (msg, parm);
 }
 
-static Widget *mousedispatch_new (int y, int x, int yl, int xl)
+static Widget *
+mousedispatch_new (int y, int x, int yl, int xl)
 {
     Widget *w = g_new (Widget, 1);
 
-    init_widget (w, y, x, yl, xl,
-		 (callback_fn) md_callback, 0, (mouse_h)  help_event, NULL);
+    init_widget (w, y, x, yl, xl, (callback_fn) md_callback,
+		 (mouse_h) help_event);
 
     return w;
 }
 
-static int help_handle_key (struct Dlg_head *h, int c)
+static cb_ret_t
+help_handle_key (struct Dlg_head *h, int c)
 {
-    char *new_item;
+    const char *new_item;
 
     if (c != KEY_UP && c != KEY_DOWN &&
-	check_movement_keys (c, 1, help_lines, currentpoint,
+	check_movement_keys (c, help_lines, currentpoint,
 			     (movefn) move_backward2,
 			     (movefn) move_forward2,
 			     (movefn) move_to_top,
@@ -719,25 +724,28 @@ static int help_handle_key (struct Dlg_head *h, int c)
 	break;
 
     default:
-	return 0;
+	return MSG_NOT_HANDLED;
 	    
     }
-    help_callback (h, 0, DLG_DRAW);
-    return 1;
+    help_callback (h, DLG_DRAW, 0);
+    return MSG_HANDLED;
 }
 
-static int help_callback (struct Dlg_head *h, int id, int msg)
+static cb_ret_t
+help_callback (struct Dlg_head *h, dlg_msg_t msg, int parm)
 {
-    switch (msg){
+    switch (msg) {
     case DLG_DRAW:
 	common_dialog_repaint (h);
 	help_show (h, currentpoint);
-	break;
+	return MSG_HANDLED;
 
     case DLG_KEY:
-	return help_handle_key (h, id);
+	return help_handle_key (h, parm);
+
+    default:
+	return default_dlg_callback (h, msg, parm);
     }
-    return 0;
 }
 
 static void
@@ -748,11 +756,11 @@ interactive_display_finish (void)
 }
 
 void
-interactive_display (char *filename, char *node)
+interactive_display (const char *filename, const char *node)
 {
     WButtonBar *help_bar;
     Widget *md;
-    char *hlpfile = filename;
+    char *hlpfile = NULL;
 
     if (filename)
 	data = load_file (filename);
@@ -760,7 +768,7 @@ interactive_display (char *filename, char *node)
 	data = load_mc_home_file ("mc.hlp", &hlpfile);
 
     if (data == NULL) {
-	message (1, MSG_ERROR, _(" Cannot open file %s \n %s "), hlpfile,
+	message (1, MSG_ERROR, _(" Cannot open file %s \n %s "), filename ? filename : hlpfile,
 		 unix_error_string (errno));
     }
 
@@ -807,8 +815,8 @@ interactive_display (char *filename, char *node)
 
     md = mousedispatch_new (1, 1, help_lines, HELP_WINDOW_WIDTH - 2);
 
-    add_widget (whelp, help_bar);
     add_widget (whelp, md);
+    add_widget (whelp, help_bar);
 
     define_label_data (whelp, 1, _("Help"), (buttonbarfn) help_help_cmd,
 		       whelp);
