@@ -11,8 +11,6 @@
                1997 Norbert Warmuth.
 	       2000 Pavel Machek
    
-   $Id: undelfs.c,v 1.14 2000/08/22 22:50:20 proskin Exp $
- 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License
    as published by the Free Software Foundation; either version 2 of
@@ -25,7 +23,7 @@
 
    You should have received a copy of the GNU Library General Public
    License along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Assumptions:
  *
@@ -39,22 +37,26 @@
 #include <errno.h>
 
 #include <stdio.h>
-#include <fcntl.h>
 #include <stdlib.h>
-/* asm/types.h defines its own umode_t :-( */
+
+#ifdef HAVE_EXT2FS_EXT2_FS_H
+#include <ext2fs/ext2_fs.h>
+#else
+/* asm/types.h defines its own umode_t */
 #undef umode_t
 #include <linux/ext2_fs.h>
+#endif
+
 #include <ext2fs/ext2fs.h>
 #include <ctype.h>
 
 #include "utilvfs.h"
 
 #include "vfs.h"
-
-void print_vfs_message(char *, ...);
+#include "vfs-impl.h"
 
 struct deleted_info {
-    ino_t   ino;
+    ext2_ino_t   ino;
     unsigned short mode;
     unsigned short uid;
     unsigned short gid;
@@ -65,7 +67,7 @@ struct deleted_info {
 };
 
 struct lsdel_struct {
-    ino_t                   inode;
+    ext2_ino_t                   inode;
     int                     num_blocks;
     int                     free_blocks;
     int                     bad_blocks;
@@ -81,6 +83,7 @@ static char *block_buf;
 static char *undelfserr = N_(" undelfs: error ");
 static int readdir_ptr;
 static int undelfs_usage;
+static struct vfs_class vfs_undelfs_ops;
 
 /* To generate the . and .. entries use -2 */
 #define READDIR_PTR_INIT 0
@@ -91,28 +94,25 @@ undelfs_shutdown (void)
     if (fs)
 	ext2fs_close (fs);
     fs = NULL;
-    if (ext2_fname)
-	g_free (ext2_fname);
+    g_free (ext2_fname);
     ext2_fname = NULL;
-    if (delarray)
-        g_free (delarray);
+    g_free (delarray);
     delarray = NULL;
-    if (block_buf)
-	g_free (block_buf);
+    g_free (block_buf);
     block_buf = NULL;
 }
 
 static void
-undelfs_get_path (char *dirname, char **ext2_fname, char **file)
+undelfs_get_path (const char *dirname, char **fsname, char **file)
 {
-    char *p;
+    const char *p;
 
     /* To look like filesystem, we have virtual directories
        /#undel:XXX, which have no subdirectories. XXX is replaced with
        hda5, sdb8 etc, which is assumed to live under /dev. 
                                                     -- pavel@ucw.cz */
 
-    *ext2_fname = NULL;
+    *fsname = NULL;
     
     if (strncmp (dirname, "/#undel:", 8))
 	return;
@@ -134,21 +134,23 @@ undelfs_get_path (char *dirname, char **ext2_fname, char **file)
     
     while (p > dirname){
 	if (*p == '/'){
+	    char *tmp;
+
 	    *file = g_strdup (p+1);
-	    *p = 0;
-	    *ext2_fname = g_strconcat ("/dev/", dirname, NULL);
-	    *p = '/';
+	    tmp = g_strndup (dirname, p - dirname);
+	    *fsname = g_strconcat ("/dev/", tmp, (char *) NULL);
+	    g_free (tmp);
 	    return;
 	}
 	p--;
     }
     *file = g_strdup ("");
-    *ext2_fname = g_strconcat ("/dev/", dirname, NULL);
+    *fsname = g_strconcat ("/dev/", dirname, (char *) NULL);
     return;
 }
 
 static int
-lsdel_proc(ext2_filsys fs, blk_t *block_nr, int blockcnt, void *private)
+undelfs_lsdel_proc(ext2_filsys fs, blk_t *block_nr, int blockcnt, void *private)
 {
     struct lsdel_struct *lsd = (struct lsdel_struct *) private;
     
@@ -166,46 +168,50 @@ lsdel_proc(ext2_filsys fs, blk_t *block_nr, int blockcnt, void *private)
     return 0;
 }
 
-/* We don't use xmalloc, since we want to recover and not abort the program
- * if we don't have enough memory
+/*
+ * Load information about deleted files.
+ * Don't abort if there is not enough memory - load as much as we can.
  */
 static int
 undelfs_loaddel (void)
 {
     int retval, count;
-    ino_t ino;
+    ext2_ino_t ino;
     struct ext2_inode inode;
-    ext2_inode_scan   scan;
-    
+    ext2_inode_scan scan;
+
     max_delarray = 100;
     num_delarray = 0;
-    delarray = g_new (struct deleted_info, max_delarray);
+    delarray = g_try_malloc (sizeof (struct deleted_info) * max_delarray);
     if (!delarray) {
-	message_1s (1, undelfserr, _(" not enough memory "));
+	message (1, undelfserr, _(" not enough memory "));
 	return 0;
     }
-    block_buf = g_malloc (fs->blocksize * 3);
+    block_buf = g_try_malloc (fs->blocksize * 3);
     if (!block_buf) {
-	message_1s (1, undelfserr, _(" while allocating block buffer "));
+	message (1, undelfserr, _(" while allocating block buffer "));
 	goto free_delarray;
     }
-    if ((retval = ext2fs_open_inode_scan(fs, 0, &scan))){
-	message_1s1d (1, undelfserr, _(" open_inode_scan: %d "), retval);
+    if ((retval = ext2fs_open_inode_scan (fs, 0, &scan))) {
+	message (1, undelfserr, _(" open_inode_scan: %d "), retval);
 	goto free_block_buf;
     }
-    if ((retval = ext2fs_get_next_inode(scan, &ino, &inode))){
-	message_1s1d (1, undelfserr, _(" while starting inode scan %d "), retval);
+    if ((retval = ext2fs_get_next_inode (scan, &ino, &inode))) {
+	message (1, undelfserr, _(" while starting inode scan %d "),
+		 retval);
 	goto error_out;
     }
 
     count = 0;
     while (ino) {
-	if ((count++ % 1024) == 0) 
-	    print_vfs_message (_("undelfs: loading deleted files information %d inodes"), count);
+	if ((count++ % 1024) == 0)
+	    print_vfs_message (_
+			       ("undelfs: loading deleted files information %d inodes"),
+			       count);
 	if (inode.i_dtime == 0)
 	    goto next;
 
-	if (S_ISDIR(inode.i_mode))
+	if (S_ISDIR (inode.i_mode))
 	    goto next;
 
 	lsd.inode = ino;
@@ -213,22 +219,30 @@ undelfs_loaddel (void)
 	lsd.free_blocks = 0;
 	lsd.bad_blocks = 0;
 
-	retval = ext2fs_block_iterate(fs, ino, 0, block_buf,
-				      lsdel_proc, &lsd);
+	retval =
+	    ext2fs_block_iterate (fs, ino, 0, block_buf,
+				  undelfs_lsdel_proc, &lsd);
 	if (retval) {
-	    message_1s1d (1, undelfserr, _(" while calling ext2_block_iterate %d "), retval);
+	    message (1, undelfserr,
+		     _(" while calling ext2_block_iterate %d "), retval);
 	    goto next;
 	}
 	if (lsd.free_blocks && !lsd.bad_blocks) {
 	    if (num_delarray >= max_delarray) {
-		max_delarray += 50;
-		delarray = g_renew (struct deleted_info, delarray, max_delarray);
-		if (!delarray) {
-		    message_1s (1, undelfserr, _(" no more memory while reallocating array "));
+		struct deleted_info *delarray_new =
+		    g_try_realloc (delarray,
+				   sizeof (struct deleted_info) *
+				   (max_delarray + 50));
+		if (!delarray_new) {
+		    message (1, undelfserr,
+			     _
+			     (" no more memory while reallocating array "));
 		    goto error_out;
 		}
+		delarray = delarray_new;
+		max_delarray += 50;
 	    }
-	    
+
 	    delarray[num_delarray].ino = ino;
 	    delarray[num_delarray].mode = inode.i_mode;
 	    delarray[num_delarray].uid = inode.i_uid;
@@ -240,39 +254,51 @@ undelfs_loaddel (void)
 	    num_delarray++;
 	}
 
-    next:
-	retval = ext2fs_get_next_inode(scan, &ino, &inode);
+      next:
+	retval = ext2fs_get_next_inode (scan, &ino, &inode);
 	if (retval) {
-	    message_1s1d(1, undelfserr, _(" while doing inode scan %d "), retval);
+	    message (1, undelfserr, _(" while doing inode scan %d "),
+		     retval);
 	    goto error_out;
 	}
     }
     readdir_ptr = READDIR_PTR_INIT;
     ext2fs_close_inode_scan (scan);
     return 1;
-    
-error_out:    
+
+  error_out:
     ext2fs_close_inode_scan (scan);
-free_block_buf:
+  free_block_buf:
     g_free (block_buf);
     block_buf = NULL;
-free_delarray:
+  free_delarray:
     g_free (delarray);
     delarray = NULL;
     return 0;
 }
 
-void com_err (const char *str, long err_code, const char *s2, ...)
-{
-    char *cptr;
 
-    cptr = g_strdup_printf (" %s (%s: %ld) ", s2, str, err_code);
-    message_1s (1, _(" Ext2lib error "), cptr);
-    g_free (cptr);
+/*
+ * This function overrides com_err() from libcom_err library.
+ * It is used in libext2fs to report errors.
+ */
+void
+com_err (const char *whoami, long err_code, const char *fmt, ...)
+{
+    va_list ap;
+    char *str;
+
+    va_start (ap, fmt);
+    str = g_strdup_vprintf (fmt, ap);
+    va_end (ap);
+
+    message (1, _(" Ext2lib error "), " %s (%s: %ld) ", str, whoami,
+	     err_code);
+    g_free (str);
 }
 
 static void *
-undelfs_opendir (vfs *me, char *dirname)
+undelfs_opendir (struct vfs_class *me, const char *dirname)
 {
     char *file, *f;
     
@@ -294,68 +320,56 @@ undelfs_opendir (vfs *me, char *dirname)
     }
 
     if (ext2fs_open (ext2_fname, 0, 0, 0, unix_io_manager, &fs)){
-	message_2s (1, undelfserr, _(" Could not open file %s "), ext2_fname);
+	message (1, undelfserr, _(" Cannot open file %s "), ext2_fname);
 	return 0;
     }
     print_vfs_message (_("undelfs: reading inode bitmap..."));
     if (ext2fs_read_inode_bitmap (fs)){
-	message_2s (1, undelfserr,
-		 _(" Could not load inode bitmap from: \n %s \n"), ext2_fname);
+	message (1, undelfserr,
+		 _(" Cannot load inode bitmap from: \n %s \n"), ext2_fname);
 	goto quit_opendir;
     }
     print_vfs_message (_("undelfs: reading block bitmap..."));
     if (ext2fs_read_block_bitmap (fs)){
-	message_2s (1, undelfserr,
-		 _(" Could not load block bitmap from: \n %s \n"), ext2_fname);
+	message (1, undelfserr,
+		 _(" Cannot load block bitmap from: \n %s \n"), ext2_fname);
 	goto quit_opendir;
     }
     /* Now load the deleted information */
     if (!undelfs_loaddel ())
 	goto quit_opendir;
-    print_vfs_message (_("undelfs: done."));
+    print_vfs_message (_("%s: done."), me->name);
     return fs;
 quit_opendir:
-    print_vfs_message (_("undelfs: failure"));
+    print_vfs_message (_("%s: failure"), me->name);
     ext2fs_close (fs);
     fs = NULL;
     return 0;
 }
 
-/* Explanation:
- * On some operating systems (Slowaris 2 for example)
- * the d_name member is just a char long (Nice trick that break everything,
- * so we need to set up some space for the filename.
- */
-static struct {
-    struct dirent dent;
-#ifdef NEED_EXTRA_DIRENT_BUFFER
-    char extra_buffer [MC_MAXPATHLEN];
-#endif
-} undelfs_readdir_data;
 
 static void *
-undelfs_readdir (void *vfs_info)
+undelfs_readdir(void *vfs_info)
 {
-    char *dirent_dest;
-    
-    if (vfs_info != fs){
-	message_1s (1, undelfserr, _(" vfs_info is not fs! "));
+    static union vfs_dirent undelfs_readdir_data;
+    static char *const dirent_dest = undelfs_readdir_data.dent.d_name;
+
+    if (vfs_info != fs) {
+	message (1, undelfserr, _(" vfs_info is not fs! "));
 	return NULL;
     }
     if (readdir_ptr == num_delarray)
 	return NULL;
-    dirent_dest = (char *) &(undelfs_readdir_data.dent.d_name [0]);
     if (readdir_ptr < 0)
-	g_snprintf(dirent_dest, MC_MAXPATHLEN, "%s", readdir_ptr == -2 ? "." : "..");
+	strcpy (dirent_dest, readdir_ptr == -2 ? "." : "..");
     else
 	g_snprintf(dirent_dest, MC_MAXPATHLEN, "%ld:%d",
-		 (long)delarray [readdir_ptr].ino,
-		 delarray [readdir_ptr].num_blocks);
+		   (long) delarray[readdir_ptr].ino,
+		   delarray[readdir_ptr].num_blocks);
     readdir_ptr++;
-    
-#if 0
-    undelfs_readdir_data.dent.d_namlen = strlen (undelfs_readdir_data.dent.d_name);
-#endif
+
+    compute_namelen(&undelfs_readdir_data.dent);
+
     return &undelfs_readdir_data;
 }
 
@@ -372,7 +386,7 @@ typedef struct {
     int  pos;			/* file position */
     int  current;		/* used to determine current position in itereate */
     int  finished;
-    ino_t inode;
+    ext2_ino_t inode;
     int  bytes_read;
     long size;
     
@@ -383,19 +397,20 @@ typedef struct {
 
 /* We do not support lseek */
 static void *
-undelfs_open (vfs *me, char *fname, int flags, int mode)
+undelfs_open (struct vfs_class *me, const char *fname, int flags, int mode)
 {
     char *file, *f;
-    ino_t  inode, i;
+    ext2_ino_t inode, i;
     undelfs_file *p = NULL;
-    
+
     /* Only allow reads on this file system */
     undelfs_get_path (fname, &file, &f);
     if (!file)
 	return 0;
-    
-    if (!ext2_fname || strcmp (ext2_fname, file)){
-	message_1s (1, undelfserr, _(" You have to chdir to extract files first "));
+
+    if (!ext2_fname || strcmp (ext2_fname, file)) {
+	message (1, undelfserr,
+		 _(" You have to chdir to extract files first "));
 	g_free (file);
 	g_free (f);
 	return 0;
@@ -403,19 +418,19 @@ undelfs_open (vfs *me, char *fname, int flags, int mode)
     inode = atol (f);
 
     /* Search the file into delarray */
-    for (i = 0; i < num_delarray; i++){
-	if (inode != delarray [i].ino)
+    for (i = 0; i < num_delarray; i++) {
+	if (inode != delarray[i].ino)
 	    continue;
 
 	/* Found: setup all the structures needed by read */
-	p = g_new (undelfs_file, 1);
-	if (!p){
+	p = (undelfs_file *) g_try_malloc (((gsize) sizeof (undelfs_file)));
+	if (!p) {
 	    g_free (file);
 	    g_free (f);
 	    return 0;
 	}
-	p->buf = g_malloc (fs->blocksize);
-	if (!p->buf){
+	p->buf = g_try_malloc (fs->blocksize);
+	if (!p->buf) {
 	    g_free (p);
 	    g_free (file);
 	    g_free (f);
@@ -426,7 +441,7 @@ undelfs_open (vfs *me, char *fname, int flags, int mode)
 	p->f_index = i;
 	p->error_code = 0;
 	p->pos = 0;
-	p->size = delarray [i].size;
+	p->size = delarray[i].size;
     }
     g_free (file);
     g_free (f);
@@ -445,7 +460,7 @@ undelfs_close (void *vfs_info)
 }
 
 static int
-dump_read(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private)
+undelfs_dump_read(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private)
 {
     int     copy_count;
     undelfs_file *p = (undelfs_file *) private;
@@ -516,9 +531,9 @@ undelfs_read (void *vfs_info, char *buffer, int count)
 	p->count = p->size - p->pos;
     }
     retval = ext2fs_block_iterate(fs, p->inode, 0, NULL,
-				  dump_read, p);
+				  undelfs_dump_read, p);
     if (retval){
-	message_1s (1, undelfserr, _(" while iterating over blocks "));
+	message (1, undelfserr, _(" while iterating over blocks "));
 	return -1;
     }
     if (p->error_code && !p->finished)
@@ -530,7 +545,7 @@ undelfs_read (void *vfs_info, char *buffer, int count)
 static long
 undelfs_getindex (char *path)
 {
-    ino_t inode = atol (path);
+    ext2_ino_t inode = atol (path);
     int i;
 
     for (i = 0; i < num_delarray; i++){
@@ -541,7 +556,7 @@ undelfs_getindex (char *path)
 }
 
 static int
-do_stat (int inode_index, struct stat *buf)
+undelfs_stat_int (int inode_index, struct stat *buf)
 {
     buf->st_dev   = 0;
     buf->st_ino   = delarray [inode_index].ino;
@@ -557,7 +572,7 @@ do_stat (int inode_index, struct stat *buf)
 }
 
 static int
-undelfs_lstat(vfs *me, char *path, struct stat *buf)
+undelfs_lstat (struct vfs_class *me, const char *path, struct stat *buf)
 {
     int inode_index;
     char *file, *f;
@@ -579,7 +594,7 @@ undelfs_lstat(vfs *me, char *path, struct stat *buf)
     }
 	
     if (!ext2_fname || strcmp (ext2_fname, file)){
-	message_1s (1, undelfserr, _(" You have to chdir to extract files first "));
+	message (1, undelfserr, _(" You have to chdir to extract files first "));
 	g_free (file);
 	g_free (f);
 	return 0;
@@ -591,26 +606,21 @@ undelfs_lstat(vfs *me, char *path, struct stat *buf)
     if (inode_index == -1)
 	return -1;
 
-    return do_stat (inode_index, buf);
-}
-
-static int
-undelfs_stat(vfs *me, char *path, struct stat *buf)
-{
-    return undelfs_lstat (me, path, buf);
+    return undelfs_stat_int (inode_index, buf);
 }
 
+#define undelfs_stat undelfs_lstat
 
 static int
 undelfs_fstat (void *vfs_info, struct stat *buf)
 {
     undelfs_file *p = vfs_info;
     
-    return do_stat (p->f_index, buf);
+    return undelfs_stat_int (p->f_index, buf);
 }
 
 static int
-undelfs_chdir(vfs *me, char *path)
+undelfs_chdir(struct vfs_class *me, const char *path)
 {
     char *file, *f;
     int fd;
@@ -623,7 +633,7 @@ undelfs_chdir(vfs *me, char *path)
     /* this could be fixed by making an ext2fs io manager to use */
     /* our vfs, but that is left as an excercise for the reader */
     if ((fd = open (file, O_RDONLY)) == -1){
-	message_2s (1, undelfserr, _(" Could not open file: %s "), file);
+	message (1, undelfserr, _(" Cannot open file %s "), file);
 	g_free (f);
 	g_free (file);
 	return -1;
@@ -642,19 +652,17 @@ undelfs_lseek(void *vfs_info, off_t offset, int whence)
 }
 
 static vfsid
-undelfs_getid(vfs *me, char *path, struct vfs_stamping **parent)
+undelfs_getid (struct vfs_class *me, const char *path)
 {
-    char *ext2_fname, *file;
+    char *fname, *fsname;
 
-    /* We run only on the local fs */
-    *parent = NULL;
-    undelfs_get_path (path, &ext2_fname, &file);
+    undelfs_get_path (path, &fsname, &fname);
 
-    if (!ext2_fname)
-        return (vfsid) -1;
-    g_free (ext2_fname);
-    g_free (file);
-    return (vfsid)0;
+    if (!fsname)
+	return NULL;
+    g_free (fname);
+    g_free (fsname);
+    return (vfsid) fs;
 }
 
 static int
@@ -671,7 +679,7 @@ undelfs_free(vfsid id)
 
 #ifdef	ENABLE_NLS
 static int
-undelfs_init(vfs *me) {
+undelfs_init(struct vfs_class *me) {
     undelfserr = _(undelfserr);
     return 1;
 }
@@ -679,59 +687,25 @@ undelfs_init(vfs *me) {
 #define undelfs_init	NULL
 #endif
 
-vfs vfs_undelfs_ops = {
-    NULL,	/* This is place of next pointer */
-    N_("Undelete filesystem for ext2"),
-    0,	/* flags */
-    "undel:",	/* prefix */
-    NULL,	/* data */
-    0,		/* errno */
-    undelfs_init,
-    NULL,
-    NULL,	/* fill_names */
-    NULL,
-
-    undelfs_open,
-    undelfs_close,
-    undelfs_read,
-    NULL,
-    
-    undelfs_opendir,
-    undelfs_readdir,
-    undelfs_closedir,
-    NULL,
-    NULL,
-
-    undelfs_stat,
-    undelfs_lstat,
-    undelfs_fstat,
-
-    NULL,
-    NULL,
-    NULL,
-
-    NULL,	/* readlink */
-    NULL,
-    NULL,
-    NULL,
-
-    NULL,
-    undelfs_chdir,
-    NULL,
-    undelfs_lseek, 
-    NULL,
-    
-    undelfs_getid,
-    undelfs_nothingisopen,
-    undelfs_free,
-    
-    NULL,	/* get_local_copy */
-    NULL,
-
-    NULL,
-    NULL,
-    NULL,
-    NULL
-
-MMAPNULL
-};
+void
+init_undelfs (void)
+{
+    vfs_undelfs_ops.name = "undelfs";
+    vfs_undelfs_ops.prefix = "undel:";
+    vfs_undelfs_ops.init = undelfs_init;
+    vfs_undelfs_ops.open = undelfs_open;
+    vfs_undelfs_ops.close = undelfs_close;
+    vfs_undelfs_ops.read = undelfs_read;
+    vfs_undelfs_ops.opendir = undelfs_opendir;
+    vfs_undelfs_ops.readdir = undelfs_readdir;
+    vfs_undelfs_ops.closedir = undelfs_closedir;
+    vfs_undelfs_ops.stat = undelfs_stat;
+    vfs_undelfs_ops.lstat = undelfs_lstat;
+    vfs_undelfs_ops.fstat = undelfs_fstat;
+    vfs_undelfs_ops.chdir = undelfs_chdir;
+    vfs_undelfs_ops.lseek = undelfs_lseek;
+    vfs_undelfs_ops.getid = undelfs_getid;
+    vfs_undelfs_ops.nothingisopen = undelfs_nothingisopen;
+    vfs_undelfs_ops.free = undelfs_free;
+    vfs_register_class (&vfs_undelfs_ops);
+}
